#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from selinonTestCase import SelinonTestCase
from selinon import SystemState


class TestSystemState(SelinonTestCase):
    def test_simple1(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     Task2
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Run without change at first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(system_state.node_args)
        self.assertIsNotNone(retry)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(len(state_dict['active_nodes']), 1)
        self.assertEqual(len(state_dict['finished_nodes']), 1)
        self.assertEqual(len(state_dict['waiting_edges']), 1)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertNotIn('Task2', state_dict['finished_nodes'].keys())
        # the result of Task1 should be propagated to Task2
        self.assertEqual(system_state.node_args, 1)

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2, 1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        # All tasks done
        self.assertIsNone(retry)
        self.assertEqual(len(state_dict['active_nodes']), 0)
        self.assertEqual(len(state_dict['finished_nodes']), 2)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())

    def test_2_to_1(self):
        #
        # flow1:
        #
        #    Task1      Task2
        #      |          |
        #       ----------
        #           |
        #         Task3
        #
        edge_table = {
                      'flow1': [{'from': ['Task1', 'Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                                {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Run without change at first
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task2 has finished
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(state_dict['active_nodes'][0]['name'], 'Task1')
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertNotIn('Task1', state_dict['finished_nodes'].keys())
        self.assertNotIn('Task3', state_dict['finished_nodes'].keys())
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has finished
        self.set_finished(task1, "some result of task1 that shouldn't be propagated")

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(state_dict['active_nodes'][0]['name'], 'Task3')
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertNotIn('Task3', state_dict['finished_nodes'].keys())
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # No change so far
        task3 = self.get_task('Task3')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(state_dict['active_nodes'][0]['name'], 'Task3')
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertNotIn('Task3', state_dict['finished_nodes'].keys())
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task 3 has finished
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task3', state_dict['finished_nodes'].keys())

    def test_2_to_1_separate(self):
        #
        # flow1:
        #
        #    Task1   Task2
        #      \      /
        #       \    /
        #        \  /
        #        Task3
        #
        # Note:
        #   Each time only one task instance finishes
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])

        # Run without change at first
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])

        # Task1 has finished
        self.set_finished(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 2)
        self.assertIn(state_dict['active_nodes'][0]['name'], ['Task2', 'Task3'])
        self.assertIn(state_dict['active_nodes'][1]['name'], ['Task2', 'Task3'])
        self.assertIn('Task1', state_dict['finished_nodes'].keys())

        # No change so far
        task3_0 = self.get_task('Task3')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 2)
        self.assertIn(state_dict['active_nodes'][0]['name'], ['Task2', 'Task3'])
        self.assertIn(state_dict['active_nodes'][1]['name'], ['Task2', 'Task3'])
        self.assertIn('Task1', state_dict['finished_nodes'].keys())

        # Task2 has finished
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 2)
        self.assertEqual(state_dict['active_nodes'][0]['name'], 'Task3')
        self.assertEqual(state_dict['active_nodes'][1]['name'], 'Task3')
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())

        # Again, no change
        task3_1 = self.get_task('Task3')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 2)
        self.assertEqual(state_dict['active_nodes'][0]['name'], 'Task3')
        self.assertEqual(state_dict['active_nodes'][1]['name'], 'Task3')
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())

        # Now the second instance of Task3 has finished
        self.set_finished(task3_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 1)
        self.assertEqual(state_dict['active_nodes'][0]['name'], 'Task3')
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task3', state_dict['finished_nodes'].keys())

        # The last task - the instance of Task3 has just finished
        self.set_finished(task3_0)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task3']), 2)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())

    def test_2_to_1_separate_2(self):
        #
        # flow1:
        #
        #    Task1   Task2
        #      \      /
        #       \    /
        #        \  /
        #        Task3
        #
        # Note:
        #   Task1 and Task2 finish at the same time
        #   Both instances of Task3 finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])

        # Task1, Task2 has finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        self.set_finished(task1)
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 2)
        self.assertEqual(state_dict['active_nodes'][0]['name'], 'Task3')
        self.assertEqual(state_dict['active_nodes'][1]['name'], 'Task3')
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())

        # Now both Task3 instances finished
        task3_0 = self.get_task('Task3', 0)
        task3_1 = self.get_task('Task3', 1)
        self.set_finished(task3_0)
        self.set_finished(task3_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task3']), 2)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())

    def test_3_to_1(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #             Task4
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])

        # Task1 has finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 2)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertNotIn('Task4', self.instantiated_tasks)

        # Task2 and Task3 have finished
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 1)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task3', state_dict['finished_nodes'].keys())
        self.assertIn('Task4', self.instantiated_tasks)

        # Wait for Task4
        task4 = self.get_task('Task4')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('active_nodes')), 1)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task3', state_dict['finished_nodes'].keys())
        self.assertIn('Task4', self.instantiated_tasks)

        # Finish the flow
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)
        self.assertEqual(len(state_dict.get('finished_nodes')), 4)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task3', state_dict['finished_nodes'].keys())
        self.assertIn('Task4', state_dict['finished_nodes'].keys())

    def test_3_to_1_fail(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #               X
        #               |
        #             Task4
        #
        # Note:
        #   The condition will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4'], 'condition': self.cond_false},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])

        # Task1, Task2 and Task3 have finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task1)
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertIn('Task1', state_dict['finished_nodes'].keys())
        self.assertIn('Task2', state_dict['finished_nodes'].keys())
        self.assertIn('Task3', state_dict['finished_nodes'].keys())
        self.assertNotIn('Task4', state_dict['finished_nodes'].keys())

    def test_1_to_2_separate(self):
        #
        # flow1:
        #
        #         Task1
        #           |
        #       ----------
        #      |          |
        #    Task2      Task3
        #
        # Note:
        #   Task3 finishes before Task2
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change so far
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task2 has finished
        task2= self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)

    def test_1_to_2(self):
        #
        # flow1:
        #
        #         Task1
        #           |
        #       ----------
        #      |          |
        #    Task2      Task3
        #
        # Note:
        #   Task2 and Task3 finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task2, Task3 have finished
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)

    def test_1_to_3_separate(self):
        #
        # flow1:
        #
        #             Task1
        #               |
        #       ----------------
        #      |        |       |
        #    Task2    Task3   Task4
        #
        # Note:
        #   Child tasks do not finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3', 'Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task1 has finished
        task1_result = {'foo': ['bar']}
        task1 = self.get_task('Task1')
        self.set_finished(task1, task1_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 3)

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task4 has finished
        task4 = self.get_task('Task4')
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 4)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)

    def test_1_to_3(self):
        #
        # flow1:
        #
        #             Task1
        #               |
        #       ----------------
        #      |        |       |
        #    Task2    Task3   Task4
        #
        # Note:
        #   Child tasks finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3', 'Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task1 has finished
        task1_result = {'foo': ['bar']}
        task1 = self.get_task('Task1')
        self.set_finished(task1, task1_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 3)

        # Task2 has finished
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        task4 = self.get_task('Task4')
        self.set_finished(task2)
        self.set_finished(task3)
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 4)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)

    def test_3_to_3(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #               |
        #       ----------------
        #      |        |       |
        #    Task4    Task5   Task6
        #
        # Note:
        #   Task1 and Task2 finish at the same time, Task3 afterwards. Task4, Task5 and Task6 finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4', 'Task5', 'Task6'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertNotIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 3)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertNotIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 3)

        # Task1, Task2 have finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        self.set_finished(task1)
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertNotIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertNotIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertIn('Task5', self.instantiated_tasks)
        self.assertIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 3)

        # Task4, Task5, Task6 have finished
        task4 = self.get_task('Task4')
        task5 = self.get_task('Task5')
        task6 = self.get_task('Task6')
        self.set_finished(task4)
        self.set_finished(task5)
        self.set_finished(task6)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertIn('Task5', self.instantiated_tasks)
        self.assertIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 6)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)

    def test_3_to_3_2(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #               |
        #       ----------------
        #      |        |       |
        #    Task4    Task5   Task6
        #
        # Note:
        #   Task1, Task2 and Task3 finish at the same time; Task4, Task5 finish at the same time, then Task6
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4', 'Task5', 'Task6'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertNotIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 3)

        # Task1, Task2, Task3 have finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task1)
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertIn('Task5', self.instantiated_tasks)
        self.assertIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 3)

        # Task4 has finished
        task4 = self.get_task('Task4')
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertIn('Task5', self.instantiated_tasks)
        self.assertIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 4)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task5, Task6 have finished
        task5 = self.get_task('Task5')
        task6 = self.get_task('Task6')
        self.set_finished(task5)
        self.set_finished(task6)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertIn('Task5', self.instantiated_tasks)
        self.assertIn('Task6', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 6)
        self.assertEqual(len(state_dict.get('active_nodes')), 0)

    def test_1_to_2_recursive(self):
        #
        # flow1:
        #
        #         Task1  <------
        #           |           |
        #       ----------      |
        #      |          |     |
        #    Task2      Task3 --
        #
        # Note:
        #   Task2 finishes before Task3 in the first iteration
        #   Task3 finishes before Task2 in the second iteration
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change so far
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        task1 = self.get_task('Task1')
        self.set_finished(task1)
        # There was no Result set for Task1
        with self.assertRaises(KeyError):
            system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
            system_state.update()

        # Task1 has finished
        task1 = self.get_task('Task1')
        task1_result = "some task result"
        self.set_finished(task1, task1_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # The next iteration
        # No change first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # The second instance of Task1 has finished
        task1_1 = self.get_task('Task1')
        self.set_finished(task1_1)
        # We do not set task result, since it should be propagated from the first run of Task1

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        # There should be a second instance of Task2 and Task3 spawned as well
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task2')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task3')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task1']), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task2')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task3')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task1']), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task3 has finished
        task3_1 = self.get_task('Task3')
        self.set_finished(task3_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 3)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task2')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task3')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task1']), 2)
        self.assertEqual(len(state_dict['finished_nodes']['Task2']), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 3)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task2')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task3')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task1']), 2)
        self.assertEqual(len(state_dict['finished_nodes']['Task2']), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task2 has finished
        task2_1 = self.get_task('Task2')
        self.set_finished(task2_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 3)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task2')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task3')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task1']), 2)
        self.assertEqual(len(state_dict['finished_nodes']['Task2']), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Once more with no change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task1')), 3)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task2')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task3')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task1']), 2)
        self.assertEqual(len(state_dict['finished_nodes']['Task2']), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

    def test_2_to_1_recursive(self):
        #
        # flow1:
        #
        #    Task1      Task2 <-
        #      |          |     |
        #       ----------      |
        #           |           |
        #         Task3   ------
        #
        # Note:
        #   Task1 finishes before Task2 in the first iteration
        #   Task2 finishes before Task1 in the second iteration
        #
        #   Also note that Task3 is run:
        #      With Task1_1 from the first iteration and Task2 from the first iteration
        #      With Task1_1 from the first iteration and Task2 from the second iteration
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 2)

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change so far
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Newly spawned Task2 has finished, Task3 should be spawned
        task2_1 = self.get_task('Task2')
        self.set_finished(task2_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task2')), 2)
        self.assertEqual(len(self.get_task_instance.task_by_name('Task3')), 2)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 3)
        self.assertEqual(len(state_dict['finished_nodes']['Task2']), 2)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        task3_1 = self.get_task('Task3')

        # Check parent tasks of Task3 from the first iteration and Task3_1 from the second iteration
        self.assertEqual(task3_1.parent['Task1'], task3.parent['Task1'])
        self.assertEqual(task3_1.parent['Task1'], task1.task_id)

        self.assertNotEqual(task3_1.parent['Task2'], task3.parent['Task2'])
        self.assertEqual(task3.parent['Task2'], task2.task_id)
        self.assertEqual(task3_1.parent['Task2'], task2_1.task_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################
"""

List of events that can be traced:

+----------------------------------------+-------------------------+------------------------------------+
| Logged event                           |  Event name             | msg_dict.keys()                    |
+========================================+=========================+====================================+
| Signalize Dispatcher run by Celery,    |                         | flow_name, dispatcher_id, args,    |
| each time a Dispatcher is              | `DISPATCHER_WAKEUP`     | retry, state                       |
| started/retried                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize a flow start - called when   |                         |                                    |
| starting nodes are run                 | `FLOW_START`            | flow_name, dispatcher_id, args     |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize a task scheduling by         | `TASK_SCHEDULE`         | flow_name, task_name, task_id,     |
| Dispatcher                             |                         | parent, args                       |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize task start by Celery         | `TASK_START`            | flow_name, task_name, task_id,     |
|                                        |                         | parent, args                       |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize subflow start by Dispatcher  | `SUBFLOW_SCHEDULE`      | flow_name, dispatcher_id,          |
|                                        |                         | child_flow_name, args, parent      |
|                                        |                         | child_dispatcher_id,               |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize end of task from             | `TASK_END`              | flow_name, task_name, task_id,     |
| SelinonTaskEnvelope                    |                         | parent, args, storage              |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize a node success from          | `NODE_SUCCESSFUL`       | flow_name, dispatcher_id,          |
| Dispatcher                             |                         | node_name, node_id                 |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize that the result of task was  | `TASK_DISCARD_RESULT`   | flow_name, task_name, task_id,     |
| discarded                              |                         | parent, args, result               |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize task failure from            | `TASK_FAILURE`          | flow_name, task_name, task_id,     |
| SelinonTaskEnvelope                    |                         | parent, args, what, retried_count  |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize task retry                   | `TASK_RETRY`            | flow_name, task_name, task_id,     |
|                                        |                         | parent, args, what, retried_count  |
|                                        |                         | max_retry, retry_countdown         |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize when a flow ends because of  | `FLOW_FAILURE`          | flow_name, dispatcher_id, what     |
| error in nodes without fallback        |                         |                                    |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize unexpected dispatcher failure| `DISPATCHER_FAILURE`    | flow_name, dispatcher_id, what     |
| this should not occur (e.g. bug,       |                         | state, node_args, parent, finished |
| database connection error, ...)        |                         | retry                              |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize a node failure from          | `NODE_FAILURE`          | flow_name, dispatcher_id,          |
| Dispatcher                             |                         | node_name, node_id, what           |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize fallback evaluation          | `FALLBACK_START`        | flow_name, dispatcher_id, nodes,   |
|                                        |                         | fallback                           |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize Dispatcher retry             | `DISPATCHER_RETRY`      | flow_name, dispatcher_id, retry,   |
|                                        |                         | state_dict, args                   |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signalize flow end                     | `FLOW_END`              | flow_name, dispatcher_id,          |
|                                        |                         | finished_nodes                     |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signal storage connect                 | `STORAGE_CONNECT`       | storage_name                       |
|                                        |                         |                                    |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signal storage disconnect              | `STORAGE_DISCONNECT`    | storage_name                       |
|                                        |                         |                                    |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signal storage access for reading      | `STORAGE_RETRIEVE`      | flow_name, task_name, storage      |
|                                        |                         |                                    |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+
| Signal storage access for writing      | `STORAGE_STORE`         | flow_name, task_name, task_id,     |
|                                        |                         | storage, result, record_id         |
|                                        |                         |                                    |
+----------------------------------------+-------------------------+------------------------------------+

"""

import logging
import platform


def _default_trace_func(event, msg_dict):
    """
    Default tracing function that is used for storing results - do nothing

    :param event: event that triggered trace point
    :param msg_dict: a dict holding additional trace information for event
    """
    pass


class Trace(object):
    """
    Trace system flow actions
    """
    _trace_func = _default_trace_func
    _logger = None

    DISPATCHER_WAKEUP, \
    FLOW_START, \
    TASK_SCHEDULE, \
    TASK_START, \
    SUBFLOW_SCHEDULE, \
    TASK_END, \
    NODE_SUCCESSFUL, \
    TASK_DISCARD_RESULT, \
    TASK_FAILURE, \
    TASK_RETRY, \
    FLOW_FAILURE, \
    DISPATCHER_FAILURE, \
    NODE_FAILURE, \
    FALLBACK_START, \
    DISPATCHER_RETRY, \
    FLOW_END, \
    STORAGE_CONNECT, \
    STORAGE_DISCONNECT, \
    STORAGE_RETRIEVE, \
    STORAGE_STORE = range(20)

    _event_strings = [
        'DISPATCHER_WAKEUP',
        'FLOW_START',
        'TASK_SCHEDULE',
        'TASK_START',
        'SUBFLOW_SCHEDULE',
        'TASK_END',
        'NODE_SUCCESSFUL',
        'TASK_DISCARD_RESULT',
        'TASK_FAILURE',
        'TASK_RETRY',
        'FLOW_FAILURE',
        'DISPATCHER_FAILURE',
        'NODE_FAILURE',
        'FALLBACK_START',
        'DISPATCHER_RETRY',
        'FLOW_END',
        'STORAGE_CONNECT',
        'STORAGE_DISCONNECT',
        'STORAGE_RETRIEVE',
        'STORAGE_STORE'
    ]

    def __init__(self):
        raise NotImplementedError()

    @classmethod
    def trace_by_logging(cls, level=logging.DEBUG, logger=None):
        """
        Trace by using Python's logging

        :param level: logging level
        :param logger: optional logger that should be used
        """
        prefix = 'DISPATCHER %10s' % platform.node()

        if not logger:
            logger = logging.getLogger('selinon_trace')
            formatter = logging.Formatter(prefix + ' - %(asctime)s.%(msecs)d %(levelname)s: %(message)s',
                                          datefmt="%Y-%m-%d %H:%M:%S")
            sh = logging.StreamHandler()
            sh.setFormatter(formatter)
            logger.addHandler(sh)
            logger.setLevel(level)
            # do not propagate to parent loggers so we don't get duplicate output
            logger.propagate = False

        cls._logger = logger
        cls._trace_func = cls.logging_trace_func

    @classmethod
    def trace_by_func(cls, func):
        """
        Trace by a custom function

        :param func: function with a one single argument
        """
        cls._trace_func = func

    @classmethod
    def log(cls, event, msg_dict):
        """
        Trace work

        :param event: tracing event
        :param msg_dict: message to be printed
        """
        cls._trace_func(event, msg_dict)

    @classmethod
    def event2str(cls, event):
        """
        Translate event to it's string representation

        :param event: event
        :return: string representation of event
        """
        # We could make a hack here in order to have O(1), but let's keep it this way
        return cls._event_strings[event]

    @classmethod
    def logging_trace_func(cls, event, msg_dict):
        """
        Trace to Python's logging facilities

        :param event: event that triggered trace point
        :param msg_dict: a dict holding additional trace information for event
        """
        message = "%s: %s" % (cls.event2str(event), msg_dict)

        if event == Trace.NODE_FAILURE or event == Trace.DISPATCHER_FAILURE or event == Trace.TASK_DISCARD_RESULT:
            return Trace._logger.warn(message)
        else:
            return Trace._logger.info(message)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from selinonTestCase import SelinonTestCase
from selinon import SystemState


class TestFlow(SelinonTestCase):
    def test_simple_flow(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     flow2
        #       |
        #       |
        #     Task2
        #
        # Note:
        #    Result of Task1 is not propagated to flow2, parent are not propagated as well
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': ['flow2'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': []
        }
        queues = {
            'Task1': 'mytask1queue',
            'Task2': 'mytask2queue',
        }
        dispatcher_queue = 'mydispatcher'
        self.init(edge_table, task_queues=queues, dispatcher_queue=dispatcher_queue)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('flow2', self.instantiated_flows)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # check task's queue configuration
        task1 = self.get_task('Task1')
        self.assertEqual(task1.queue, queues.get(task1.task_name))

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('flow2', self.instantiated_flows)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has finished
        task1_result = [1, 2, 3, 4]
        self.set_finished(task1, task1_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        flow2 = self.get_flow('flow2')

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIsNone(flow2.node_args)
        self.assertEqual(flow2.flow_name, 'flow2')
        self.assertIsNone(flow2.parent)
        self.assertIsNone(flow2.state)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)
        self.assertEqual(len(state_dict['finished_nodes']), 1)
        self.assertEqual(len(state_dict['active_nodes']), 1)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIsNone(flow2.node_args)
        self.assertEqual(flow2.flow_name, 'flow2')
        self.assertIsNone(flow2.parent)
        self.assertIsNone(flow2.state)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)
        self.assertEqual(len(state_dict['finished_nodes']), 1)
        self.assertEqual(len(state_dict['active_nodes']), 1)

        # check flow queue propagation
        flow2 = self.get_flow('flow2')
        self.assertEqual(flow2.queue, dispatcher_queue)

        # flow2 has finished
        self.set_finished(flow2, {'TaskSubflow': [1]})

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        flow2 = self.get_flow('flow2')

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIsNone(flow2.node_args)
        self.assertEqual(flow2.flow_name, 'flow2')
        self.assertIsNone(flow2.parent)
        self.assertIsNone(flow2.state)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIsNone(flow2.node_args)
        self.assertIsNone(flow2.parent)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertEqual(state_dict['waiting_edges'][0], 0)
        self.assertEqual(len(state_dict['finished_nodes']), 2)
        self.assertEqual(len(state_dict['active_nodes']), 1)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertEqual(state_dict['waiting_edges'][0], 0)
        self.assertEqual(len(state_dict['finished_nodes']), 2)
        self.assertEqual(len(state_dict['active_nodes']), 1)

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertEqual(system_state.node_args, task1_result)
        self.assertEqual(task2.node_args, task1_result)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertEqual(state_dict['waiting_edges'][0], 0)
        self.assertEqual(len(state_dict['finished_nodes']), 3)
        self.assertEqual(len(state_dict['active_nodes']), 0)

    def test_propagate_finished(self):
        #
        # flow1:
        #
        #     Task1       Task2
        #       |           |
        #     flow2         |
        #       |           |
        #        -----------
        #             |
        #           TaskX
        #
        # flow2:
        #    Not run explicitly, but result finished_nodes is:
        #         {'flow3': [<flow3-id>], 'Task2': [<task2-id1>], 'Task3': [<task3-id1>, <task3-id2>]}
        # flow3
        #    Not run explicitly, but result finished_nodes is:
        #         {'Task2': [<task2-id2>], 'Task4': [<task4-id1>, <task4-id2>]}
        #
        # We are propagating finished, so we should inspect parent
        #
        edge_table = {
            'flow1': [{'from': ['flow2', 'Task2'], 'to': ['TaskX'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': []
        }
        self.init(edge_table, propagate_finished={'flow1': True})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('flow2', self.instantiated_flows)

        # Task1 and Task2 have finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, None)

        task2 = self.get_task('Task2')
        self.set_finished(task2, None)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIn('flow2', self.instantiated_flows)

        # Create flow3 manually
        flow3 = self.get_task_instance('flow3', None, None, None, None, None)

        flow2 = self.get_flow('flow2')
        self.set_finished(flow2)
        flow2_result = {'finished_nodes': {'flow3': [flow3.task_id],
                                           'Task2': ['<task2-id1>'],
                                           'Task3': ['<task3-id1>', '<task3-id2>']},
                        'failed_nodes': {}
                        }
        self.set_finished(flow2, flow2_result)

        flow3_result = {'finished_nodes':
                            {'Task2': ['<task2-id2>'], 'Task4': ['<task4-id1>', '<task4-id2>']},
                        'failed_nodes': {}
                        }
        self.set_finished(flow3, flow3_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        retry = system_state.update()

        self.assertIsNotNone(retry)

        task_x = self.get_task('TaskX')

        self.assertIn('flow2', task_x.parent)

        task_x_parent = {'Task2': task2.task_id,
                         'flow2': {'Task2': ['<task2-id1>'],
                                   'Task3': ['<task3-id1>', '<task3-id2>'],
                                   'flow3': {'Task2': ['<task2-id2>'],
                                             'Task4': ['<task4-id1>', '<task4-id2>'],
                                             }
                                   }
                         }
        self.assertEqual(task_x.parent, task_x_parent)

    def test_propagate_node_args_true(self):
        #
        # flow1:
        #
        #     flow2
        #
        # Note:
        #    Arguments are propagated to flow2
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table, propagate_node_args={'flow1': True})

        node_args = {'foo': 'bar'}
        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('flow2', self.instantiated_flows)

        flow2 = self.get_flow('flow2')
        self.assertEqual(flow2.node_args, node_args)

    def test_propagate_node_args_flow_name(self):
        #
        # flow1:
        #
        #     flow2  flow3
        #
        # Note:
        #    Arguments are propagated to flow2 but not to flow3
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2', 'flow3'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': []
        }
        self.init(edge_table, propagate_node_args={'flow1': ['flow2']})

        node_args = {'foo': 'bar'}
        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('flow3', self.instantiated_flows)

        flow2 = self.get_flow('flow2')
        self.assertEqual(flow2.node_args, node_args)

        flow3 = self.get_flow('flow3')
        self.assertIsNone(flow3.node_args)

    def test_propagate_parent_true(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     flow2
        #
        # Note:
        #    Parent are propagated to flow2
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table, propagate_parent={'flow1': True})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertNotIn('flow2', self.instantiated_flows)
        self.assertIn('Task1', self.instantiated_tasks)

        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, 0xDEADBEEF)
        self.assertIn('flow2', self.instantiated_flows)

        flow2 = self.get_flow('flow2')

        self.assertIsNone(flow2.node_args)
        self.assertIn('Task1', flow2.parent)
        self.assertEqual(flow2.parent['Task1'], task1.task_id)

    def test_propagate_parent_flow_name(self):
        #
        # flow1:
        #         Task1
        #           |
        #        -------
        #       |       |
        #     flow2   flow3
        #
        # Note:
        #    Arguments are propagated to flow2 but not to flow3. Result of Task1 is also propagated to flow2 and flow3
        #    as node_args.
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2', 'flow3'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': []
        }
        self.init(edge_table, propagate_parent={'flow1': ['flow2']}, propagate_node_args={'flow1': True})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('flow2', self.instantiated_flows)
        self.assertNotIn('flow3', self.instantiated_flows)

        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        self.assertIsNotNone(retry)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('flow3', self.instantiated_flows)

        flow2 = self.get_flow('flow2')
        self.assertEqual(flow2.node_args, 0xDEADBEEF)
        self.assertIn('Task1', flow2.parent)
        self.assertEqual(flow2.parent['Task1'], task1.task_id)

        flow3 = self.get_flow('flow3')
        self.assertEqual(flow3.node_args, 0xDEADBEEF)
        self.assertIsNone(flow3.parent)
